#include "AMDiS.h"

using namespace std;
using namespace AMDiS;

// ===========================================================================
// ===== function definitions ================================================
// ===========================================================================

/** \brief
 * Dirichlet boundary function
 */
class G : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(G);

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  double operator()(const WorldVector<double>& x) const {
    return exp(-10.0*(x*x));
  }
};

/** \brief
 * RHS function
 */
class F : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(F);

  /** \brief
   * Constructor
   */
  F(int degree, double sigma_, double k_, int dim_)
    : AbstractFunction<double, WorldVector<double> >(degree), sigma(sigma_), k(k_), dim(dim_)
  {};

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  double operator()(const WorldVector<double>& x) const {
    double r2 = x*x;
    double ux = exp(-10.0*r2);
    double ux4 = ux*ux*ux*ux;
    return sigma*ux4 - k*(400.0*r2 - 20.0*dim)*ux;
  }

private:
  double sigma;
  double k;
  int dim;
};

/** \brief
 * Needed for zero order term.
 */
class CQPFct : public AbstractFunction<double, double>
{
public:
  MEMORY_MANAGED(CQPFct);

  CQPFct() : AbstractFunction<double, double>(3) {};

  /** \brief
   * Constructor.
   */
  CQPFct(double sigma_) : sigma(sigma_) {};

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  double operator()(const double& x) const {
    return sigma * x * x * x;
  }

private:
  double sigma;
};


// ===========================================================================
// ===== class NonLin ========================================================
// ===========================================================================

/** \brief
 * Non linear problem.
 */
class NonLinVec : public ProblemNonLinVec
{
public:
  /** \brief
   * Constructor.
   */
  NonLinVec(const std::string& name_)
    : ProblemNonLinVec(name_),
      sigma(1.0)
  {};

  /** \brief
   * Returns \ref sigma.
   */
  double getSigma() { return sigma; };

  /** \brief
   * Sets \ref sigma.
   */
  void setSigma(double sig) { sigma = sig; };

private:
  /** \brief
   * Stefan-Bolzmann constant.
   */
  double sigma;
};

// ===========================================================================
// ===== main program ========================================================
// ===========================================================================

int main(int argc, char* argv[])
{
  FUNCNAME("main");

  // ===== check for init file =====
  TEST_EXIT(argc == 2)("usage: nonlin initfile\n");

  // ===== init parameters =====
  Parameters::init(false, argv[1]);

  // ===== needed constants =====
  double k = 1.0;
  double factor1 = 4.0;
  double factor2 = -1.0;

  // ===== create and init non linear problem =====
  NonLinVec vecnonlin("vecnonlin");
  vecnonlin.initialize(INIT_ALL);
  vecnonlin.setSigma(1.0);

  // === create adapt info ===
  AdaptInfo *adaptInfo = NEW AdaptInfo("vecnonlin->adapt", 
				       vecnonlin.getNumComponents());

  // === create adapt ===
  AdaptStationary *adapt = NEW AdaptStationary("vecnonlin->adapt",
					       &vecnonlin,
					       adaptInfo);

  // ===== create operators =====
  Operator *nonLinOperator0_0 = NEW Operator(Operator::MATRIX_OPERATOR | 
					     Operator::VECTOR_OPERATOR, 
					     vecnonlin.getFESpace(0),
					     vecnonlin.getFESpace(0));
  Operator *nonLinOperator0_1 = NEW Operator(Operator::MATRIX_OPERATOR | 
					     Operator::VECTOR_OPERATOR, 
					     vecnonlin.getFESpace(1),
					     vecnonlin.getFESpace(1));

  nonLinOperator0_0->setUhOld(vecnonlin.getSolution()->getDOFVector(0));
  nonLinOperator0_1->setUhOld(vecnonlin.getSolution()->getDOFVector(1));

  nonLinOperator0_0->addZeroOrderTerm
    (NEW VecAtQP_ZOT(
		     NULL,  // use uhOld
		     NEW CQPFct(vecnonlin.getSigma())
		     )
     );

  nonLinOperator0_1->addZeroOrderTerm
    (NEW VecAtQP_ZOT(
		     NULL,  // use uhOld
		     NEW CQPFct(vecnonlin.getSigma())
		    )
     );

  vecnonlin.addMatrixOperator(nonLinOperator0_0, 0, 0, &factor1, &factor1);
  vecnonlin.addMatrixOperator(nonLinOperator0_1, 1, 1, &factor1, &factor1);
  vecnonlin.addVectorOperator(nonLinOperator0_0, 0);
  vecnonlin.addVectorOperator(nonLinOperator0_1, 1);

  Operator *nonLinOperator2_0 = NEW Operator(Operator::MATRIX_OPERATOR | 
					     Operator::VECTOR_OPERATOR,
					     vecnonlin.getFESpace(0),
					     vecnonlin.getFESpace(0));
  Operator *nonLinOperator2_1 = NEW Operator(Operator::MATRIX_OPERATOR | 
					     Operator::VECTOR_OPERATOR,
					     vecnonlin.getFESpace(1),
					     vecnonlin.getFESpace(1));

  nonLinOperator2_0->setUhOld(vecnonlin.getSolution()->getDOFVector(0));
  nonLinOperator2_1->setUhOld(vecnonlin.getSolution()->getDOFVector(1));

  nonLinOperator2_0->addSecondOrderTerm(NEW FactorLaplace_SOT(&k));
  nonLinOperator2_1->addSecondOrderTerm(NEW FactorLaplace_SOT(&k));

  vecnonlin.addMatrixOperator(nonLinOperator2_0, 0, 0);
  vecnonlin.addMatrixOperator(nonLinOperator2_1, 1, 1);
  vecnonlin.addVectorOperator(nonLinOperator2_0, 0);
  vecnonlin.addVectorOperator(nonLinOperator2_1, 1);


  int dim = vecnonlin.getMesh(0)->getDim();
  int degree;
  degree = vecnonlin.getFESpace(0)->getBasisFcts()->getDegree();

  Operator* rhsFunctionOperator_0 = 
  NEW Operator(Operator::VECTOR_OPERATOR, vecnonlin.getFESpace(0));
  rhsFunctionOperator_0->addZeroOrderTerm
    (NEW CoordsAtQP_ZOT(NEW F(degree, vecnonlin.getSigma(), k, dim)));
  
  degree = vecnonlin.getFESpace(1)->getBasisFcts()->getDegree();

  Operator* rhsFunctionOperator_1 = 
  NEW Operator(Operator::VECTOR_OPERATOR, vecnonlin.getFESpace(1));
  rhsFunctionOperator_1->addZeroOrderTerm
    (NEW CoordsAtQP_ZOT(NEW F(degree, vecnonlin.getSigma(), k, dim)));
  
  vecnonlin.addVectorOperator(rhsFunctionOperator_0, 0, &factor2, &factor2);
  vecnonlin.addVectorOperator(rhsFunctionOperator_1, 1, &factor2, &factor2);

  // ===== add boundary conditions =====
  vecnonlin.addDirichletBC(DIRICHLET, 0, NEW G);
  vecnonlin.addDirichletBC(DIRICHLET, 1, NEW G);

  // ===== start adaption loop =====
  adapt->adapt();

  // ===== print solution =====
  vecnonlin.getSolution()->getDOFVector(0)->print();
  vecnonlin.getSolution()->getDOFVector(1)->print();

  return 0;
}





